import logging
import pathlib

# Third party imports
import pandas as pd

# Local application imports
from asic.reader import FileReader

from ..metadata import FileItemInfo

logger = logging.getLogger(__name__)

dspcttos_format = {
    "type": "csv",
    "sep": ";",
    "encoding": "cp1252",
    "dt_fields": {},
    "dtype": {
        "CONTRATO": str,
        "VENDEDOR": str,
        "COMPRADOR": str,
        "TIPO": str,
        "TIPOMERC": str,
        "DESP_HORA 01": float,
        "DESP_HORA 02": float,
        "DESP_HORA 03": float,
        "DESP_HORA 04": float,
        "DESP_HORA 05": float,
        "DESP_HORA 06": float,
        "DESP_HORA 07": float,
        "DESP_HORA 08": float,
        "DESP_HORA 09": float,
        "DESP_HORA 10": float,
        "DESP_HORA 11": float,
        "DESP_HORA 12": float,
        "DESP_HORA 13": float,
        "DESP_HORA 14": float,
        "DESP_HORA 15": float,
        "DESP_HORA 16": float,
        "DESP_HORA 17": float,
        "DESP_HORA 18": float,
        "DESP_HORA 19": float,
        "DESP_HORA 20": float,
        "DESP_HORA 21": float,
        "DESP_HORA 22": float,
        "DESP_HORA 23": float,
        "DESP_HORA 24": float,
        "TRF_HORA 01": float,
        "TRF_HORA 02": float,
        "TRF_HORA 03": float,
        "TRF_HORA 04": float,
        "TRF_HORA 05": float,
        "TRF_HORA 06": float,
        "TRF_HORA 07": float,
        "TRF_HORA 08": float,
        "TRF_HORA 09": float,
        "TRF_HORA 10": float,
        "TRF_HORA 11": float,
        "TRF_HORA 12": float,
        "TRF_HORA 13": float,
        "TRF_HORA 14": float,
        "TRF_HORA 15": float,
        "TRF_HORA 16": float,
        "TRF_HORA 17": float,
        "TRF_HORA 18": float,
        "TRF_HORA 19": float,
        "TRF_HORA 20": float,
        "TRF_HORA 21": float,
        "TRF_HORA 22": float,
        "TRF_HORA 23": float,
        "TRF_HORA 24": float,
    },
}


class DSPCTTOS(FileReader):
    def __init__(self):
        return super().__init__(dspcttos_format.copy())


def dspcttos_preprocess(filepath: pathlib.Path, item: FileItemInfo) -> pd.DataFrame:
    """
    DSPCTTOS: se publica un archivo por dia.
    versiones: TX2, TXR, TXF
    COMPRADOR:
        EPSC: aqui despues va enerbit
    TIPOMERC:
        R: Regulada
    """
    dspcttos_reader = DSPCTTOS()
    total = dspcttos_reader.read(filepath)
    total["FECHA"] = f"{item.year:04d}-{item.month:02d}-{item.day:02d}"

    total = total[(total["COMPRADOR"] == item.agent) & (total["TIPOMERC"] == "R")]

    total["FECHA"] = pd.to_datetime(
        total["FECHA"],
        format="%Y-%m-%d",
    )
    total = (
        total.set_index(
            ["FECHA", "CONTRATO", "VENDEDOR", "COMPRADOR", "TIPO", "TIPOMERC"]
        )
        .stack()
        .reset_index()
    )

    total = total.rename(columns={"level_6": "NOMBRE HORA", 0: "VALOR"})
    total["CONCEPTO"] = total["NOMBRE HORA"].apply(lambda x: x.split("_")[0])
    total["HORA"] = (total["NOMBRE HORA"].str.slice(start=-2)).astype(int) - 1
    total["HORA"] = pd.to_timedelta(total["HORA"], unit="h")
    total["FECHA_HORA"] = total["FECHA"] + total["HORA"]

    total = (
        total[
            [
                "FECHA",
                "HORA",
                "FECHA_HORA",
                "CONTRATO",
                "VENDEDOR",
                "COMPRADOR",
                "TIPO",
                "TIPOMERC",
                "CONCEPTO",
                "VALOR",
            ]
        ]
        .set_index(
            [
                "FECHA",
                "HORA",
                "FECHA_HORA",
                "CONTRATO",
                "VENDEDOR",
                "COMPRADOR",
                "TIPO",
                "TIPOMERC",
                "CONCEPTO",
            ]
        )
        .unstack()
        .reset_index()
    )
    cols = [
        f"{l1}_{l0}" if l1 else l0
        for (l0, l1) in zip(
            total.columns.get_level_values(0), total.columns.get_level_values(1)
        )
    ]
    total.columns = cols

    ret_cols = [
        "FECHA_HORA",
        "CONTRATO",
        "VENDEDOR",
        "COMPRADOR",
        "TIPO",
        "TIPOMERC",
        "DESP_VALOR",
        "TRF_VALOR",
    ]
    return total[ret_cols]



import logging
import pathlib

# Third party imports
import pandas as pd

# Local application imports
from asic.reader import FileReader

from ..metadata import FileItemInfo

logger = logging.getLogger(__name__)

aenc_format = {
    "type": "csv",
    "sep": ";",
    "encoding": "cp1252",
    "dt_fields": {},
    "dtype": {
        "CODIGO SIC": str,
        "CODIGO PROPIO": str,
        "TIPO DE AGRUPACIÓN": str,
        "IMPO - EXPO": str,
        "HORA 01": float,
        "HORA 02": float,
        "HORA 03": float,
        "HORA 04": float,
        "HORA 05": float,
        "HORA 06": float,
        "HORA 07": float,
        "HORA 08": float,
        "HORA 09": float,
        "HORA 10": float,
        "HORA 11": float,
        "HORA 12": float,
        "HORA 13": float,
        "HORA 14": float,
        "HORA 15": float,
        "HORA 16": float,
        "HORA 17": float,
        "HORA 18": float,
        "HORA 19": float,
        "HORA 20": float,
        "HORA 21": float,
        "HORA 22": float,
        "HORA 23": float,
        "HORA 24": float,
    },
}


class AENC(FileReader):
    def __init__(self):
        return super().__init__(aenc_format.copy())


def aenc_preprocess(filepath: pathlib.Path, item: FileItemInfo) -> pd.DataFrame:
    """
    AENC: es un archivo diario
    versiones: TX2, TXR, TXF
    VALOR: energia calculada por ASIC para frontera en cada periodo
    """
    ance_reader = AENC()
    total = ance_reader.read(filepath)
    total["FECHA"] = f"{item.year:04d}-{item.month:02d}-{item.day:02d}"
    total["AGENTE"] = item.agent

    total["FECHA"] = pd.to_datetime(
        total["FECHA"],
        format="%Y-%m-%d",
    )
    total = (
        total.set_index(
            [
                "FECHA",
                "AGENTE",
                "CODIGO SIC",
                "CODIGO PROPIO",
                "TIPO DE AGRUPACIÓN",
                "IMPO - EXPO",
            ]
        )
        .stack()
        .reset_index()
    )
    total = total.rename(columns={"level_6": "NOMBRE HORA", 0: "VALOR"})
    total["HORA"] = (total["NOMBRE HORA"].str.slice(start=-2)).astype(int) - 1
    total["HORA"] = pd.to_timedelta(total["HORA"], unit="h")
    total["FECHA_HORA"] = total["FECHA"] + total["HORA"]

    # total = (
    #     total[
    #         ["FECHA", "NOMBRE HORA", "HORA", "FECHA_HORA", "AGENTE", "CODIGO", "VALOR"]
    #     ]
    #     .set_index(["FECHA", "NOMBRE HORA", "HORA", "FECHA_HORA", "AGENTE", "CODIGO"])
    #     .unstack()
    #     .reset_index()
    # )
    # cols = [
    #     f"{l1}_{l0}" if l1 else l0
    #     for (l0, l1) in zip(
    #         total.columns.get_level_values(0), total.columns.get_level_values(1)
    #     )
    # ]
    # total.columns = cols
    return_cols = [
        "FECHA_HORA",
        "AGENTE",
        "CODIGO SIC",
        "CODIGO PROPIO",
        "TIPO DE AGRUPACIÓN",
        "IMPO - EXPO",
        "VALOR",
    ]
    return total[return_cols]


if __name__ == "__main__":
    import pathlib

    aenc_path = pathlib.Path(
        r"C:/Users/pablete/enerbit/enerbit.core.eds/files/aenc0703.Tx2"
    )
    data = AENC().read(aenc_path)
    print(data.head())
    info = FileItemInfo(
        path=aenc_path,
        code="aenc",
        year=2022,
        month=7,
        day=3,
        extension="tx2",
        version="002",
        agent="enbc",
    )
    data_proc = aenc_preprocess(aenc_path, info)
    print(data_proc.sample(5))



import datetime as dt
import ftplib
import functools
import itertools
import logging
import pathlib
import re
from typing import Iterable

import pydantic
from asic import ASIC_FILE_CONFIG, ASIC_FILE_EXTENSION_MAP

logger = logging.getLogger(__name__)

SUPPORTED_ASIC_EXTENSIONS = frozenset(ASIC_FILE_EXTENSION_MAP.keys())


class DownloadSpec(pydantic.BaseModel):
    remote: pathlib.PurePosixPath
    local: pathlib.Path

    class Config:
        arbitrary_types_allowed = True


def grab_file(ftp: ftplib.FTP, remote: pathlib.PurePath, local: pathlib.Path):
    with open(local, "wb") as dst:
        ftp.retrbinary("RETR " + str(remote), dst.write)


def grab_files(ftp: ftplib.FTP, files: Iterable[DownloadSpec]) -> None:
    for i in files:
        grab_file(ftp, i.remote, i.local)


def get_path_version(path: pathlib.PurePath) -> str:
    version = ASIC_FILE_EXTENSION_MAP[path.suffix.lower()].normalized_version
    return version


@functools.cache
def list_files_in_location(
    ftp: ftplib.FTP,
    location: str,
) -> list[pathlib.PurePath]:
    ftp.cwd(location)
    location_path = pathlib.PurePath(location)
    logger.debug(f"Listing files in location {location}")
    files_in_location = ftp.nlst()
    logger.debug(f"Total files found in location {len(files_in_location)}")
    paths_in_location = [location_path / f for f in files_in_location]
    logger.debug(f"Total paths found in location {len(paths_in_location)}")
    return paths_in_location


def fiter_files_by_pattern(
    file_list: list[pathlib.PurePath], name_pattern: str
) -> list[pathlib.PurePath]:
    filtered = [f for f in file_list if re.search(name_pattern, str(f.name))]
    return filtered


def list_supported_files(
    ftp: ftplib.FTP,
    *,
    agent: str | None = None,
    months: list[dt.date],
    extensions: list[str],
    files: list[str],
    locations: list[str],
) -> list[pathlib.PurePath]:
    file_list = []
    for month, l_template, ext in itertools.product(
        months,
        locations,
        extensions,
    ):
        try:
            remote_location = l_template.format(
                year=month.year, month=month.month, agent=agent
            )
        except Exception:
            logger.debug(
                f"Failed to build remote location with {l_template}, {month, agent}"
            )
            continue
        logger.debug(f"Listing remote location: {remote_location}")
        file_list.extend(
            list_supported_files_in_location(ftp, remote_location, files, ext)
        )

    return file_list


def list_supported_files_in_location(
    ftp: ftplib.FTP,
    location: str,
    file_codes: list[str],
    extension: str,
) -> list[pathlib.PurePath]:
    files = list_files_in_location(ftp, location)
    logger.debug(f"Total files in location {location}")

    patterns = [c.name_pattern for f, c in ASIC_FILE_CONFIG.items() if f in file_codes]

    if extension is not None:
        patterns = [combine_patterns_and_extension(p, extension) for p in patterns]

    logger.debug(f"Patterns to filter by: {patterns}")

    supported_files_in_location = []
    for p in patterns:
        logger.debug(f"Filtering {len(files)} by pattern {p}")
        new_files = fiter_files_by_pattern(files, p)
        logger.debug(f"Kept {len(new_files)} files")
        supported_files_in_location.extend(new_files)
    logger.info(f"Total files kept {len(supported_files_in_location)}")
    return supported_files_in_location


def combine_patterns_and_extension(pattern: str, extension: str) -> str:
    logger.info(f"Preparing search pattern with file extension '{extension}'")
    aux_p = pattern.split(".")
    pattern_ext = "".join(
        ["[" + a.lower() + a.upper() + "]{1}" for a in extension if a != "."]
    )
    logger.debug(f"Search pattern without extension: {pattern}")
    pattern = aux_p[0] + f".(?P<ext>{pattern_ext})"
    logger.debug(f"Search pattern with extension: {pattern}")

    return pattern


if __name__ == "__main__":
    # Mes del que se va a descargar
    # dt_format = r"%Y-%m-%d"
    # beg_month = dt.dt.datetime.strptime("2019-03-01", dt_format)  # input beg_month date
    # end_month = dt.dt.datetime.strptime("2019-07-01", dt_format)  # input end_month date
    # file_code = "adem"

    import os

    location = pathlib.PurePosixPath("/UsuariosK/ENBC/Sic/COMERCIA/2022-07/")
    with ftplib.FTP(
        host=os.environ["ASIC_FTP_HOST"],
        user=os.environ["ASIC_FTP_USER"],
        passwd=os.environ["ASIC_FTP_PASSWORD"],
    ) as ftp_aux:
        files = list_files_in_location(ftp_aux, str(location))
        for f in files:
            print(f)



import logging
import pathlib

# Third party imports
import pandas as pd

# Local application imports
from asic.reader import FileReader

from ..metadata import FileItemInfo

logger = logging.getLogger(__name__)

pep_format = {
    "type": "csv",
    "sep": ";",
    "encoding": "cp1252",
    "dt_fields": {},
    "dtype": {"AGENTE": str, "VALOR PE": float},
}


class PEP(FileReader):
    def __init__(self):
        return super().__init__(pep_format.copy())


def pep_preprocess(filepath: pathlib.Path, item: FileItemInfo) -> pd.DataFrame:
    """
    pep: se publica un archivo por día.
    versiones: TX1, TX2, TXR y TXF
    AGENTE:
      SISTEMA: el precio de escasez del sistema
    """
    pep_reader = PEP()
    total = pep_reader.read(filepath)
    total["FECHA"] = f"{item.year:04d}-{item.month:02d}-{item.day:02d}"

    total = total[total["AGENTE"] == "SISTEMA"]

    total["FECHA"] = pd.to_datetime(
        total["FECHA"],
        format="%Y-%m-%d",
    )
    return_cols = ["FECHA", "AGENTE", "VALOR PE"]
    return total[return_cols]



import logging
import pathlib

# Third party imports
import pandas as pd

# Local application imports
from asic.reader import FileReader

from ..metadata import FileItemInfo

logger = logging.getLogger(__name__)

afac_format = {
    "type": "csv",
    "sep": ";",
    "encoding": "cp1252",
    "dt_fields": {},
    "dtype": {
        "AGENTE": str,
        "PERDIDA REAL (kWh)": float,
        "DEMANDA REAL (kWh)": float,
        "GENERACION REAL (kWh)": float,
        "COMPRAS EN BOLSA (KWH)": float,
        "COMPRAS EN BOLSA ($)": float,
        "VENTAS EN BOLSA (KWH)": float,
        "VENTAS EN BOLSA ($)": float,
        "COMPRAS EN DESVIACION (KWH)": float,
        "COMPRAS EN DESVIACION ($)": float,
        "VENTAS EN DESVIACION (KWH)": float,
        "VENTAS EN DESVIACION ($)": float,
        "COMPRAS EN RECONCILIACION (KWH)": float,
        "COMPRAS EN RECONCILIACION ($)": float,
        "VENTAS EN RECONCILIACION (KWH)": float,
        "VENTAS EN RECONCILIACION ($)": float,
        "COMPRAS EN CONTRATOS (kWh)": float,
        "VENTAS EN CONTRATOS (kWh)": float,
        "COMPRAS ENERGIA EN BOLSA (KWH)": float,
        "COMPRAS ENERGIA EN BOLSA ($)": float,
        "VENTAS ENERGIA EN BOLSA (KWH)": float,
        "VENTAS ENERGIA EN BOLSA ($)": float,
        "VR CARGO POR CONFIABILIDAD ($)": float,
        "VD CARGO POR CONFIABILIDAD ($)": float,
        "NETO CXC ($)": float,
        "COMPRAS CARGO POR CONFIABILIDAD ($)": float,
        "VENTAS CARGO POR CONFIABILIDAD ($)": float,
        "COMPRAS EN BOLSA NACIONAL (KWH)": float,
        "COMPRAS EN BOLSA NACIONAL ($)": float,
        "VENTAS EN BOLSA NACIONAL (KWH)": float,
        "VENTAS EN BOLSA NACIONAL ($)": float,
        "COMPRAS EN BOLSA INTERNACIONAL (KWH)": float,
        "COMPRAS EN BOLSA INTERNACIONAL ($)": float,
        "VENTAS EN BOLSA INTERNACIONAL (KWH)": float,
        "VENTAS EN BOLSA INTERNACIONAL ($)": float,
        "SERVICIOS AGC ($)": float,
        "RESPONSABILIDAD COMERCIAL AGC (KWH)": float,
        "RESPONSABILIDAD COMERCIAL AGC ($)": float,
        "TOTAL COMPRAS ($)": float,
        "TOTAL VENTAS ($)": float,
        "VALOR A PAGAR POR SRPF ($)": float,
        "VALOR A RECIBIR POR SRPF ($)": float,
        "TOTAL RESTRICCIONES ($)": float,
        "RENTAS DE CONGESTION ($)": float,
        "RESTRICCIONES ALIVIADAS ($)": float,
        "VEBO (KWH)": float,
        "RENTAS DE CONGESTIÓN POR IMPORTACIÓN ($)": float,
        "DISTRIBUCIÓN SALDO NETO TIE EN MÉRITO ($)": float,
        "DISTRIBUCIÓN SALDO NETO TIE FUERA DE MÉRITO ($)": float,
        "COMPRAS BOLSA CON SALDO NETO TIE MÉRITO ($)": float,
        "RENDIMIENTOS FINANCIEROS POR EXPORTACIONES TIE($)": float,
        "ALIVIO POR CIOEF ($)": float,
        "COMPRAS NDC ($)": float,
        "VENTAS DESVIACIONES OEFH ($)": float,
        "COMPRAS DESVIACIONES OEFH ($)": float,
        "DEVOLUCION DINEROS DEL CARGO POR CONFIABILIDAD ($)": float,
        "COBRO DINERO CARGO POR CONFIABILIDAD ($)": float,
        "COMPRAS ARRANQUE Y PARADA ($)": float,
        "VENTAS ARRANQUE Y PARADA ($)": float,
        "Ventas por EEVE ($)": float,
        "Compras por EEVE ($)": float,
        "Restricciones por EEVE ($)": float,
        "Cobro uso respaldo($)": float,
        "Alivio restricciones RES 05/2010 ($)": float,
        "COMPRAS EN BOLSA TIES($)": float,
        " VENTAS EN BOLSA TIES ($)": float,
        " MAGNITUD EN Kwh  DE COMPRAS EN BOLSA DE TIES": float,
        " MAGNITUD EN kwh DE VENTAS EN BOLSA TIES": float,
        "ALIVIO POR EJECUCION DE GARANTIA ($)": float,
        "VALOR TOTAL EJECUCION DE GARANTIA ($)": float,
        "Alivio por VCSRCFVD($)": float,
        "VOEFV a cargo por la OEFV adquirida en la SRCFV($)": float,
        "VMOEFV a cargo al Margen del Precio MP SRCFV ($):": float,
        "Costo de Exportación ($)": float,
        "Total Costo de Exportación ($)": float,
        "Total de Generación Ideal en kWh del Agente": float,
        "Total de Holgura de AGC en kWh asignados al Agente": float,
        "Energía vendida y embalsada Asignada kWh": float,
        "VR Demanda Res 155/2014": float,
        "Alivio Asociado a la Resolución CREG 024/2015 en $": float,
        "Cobro Autogeneradores Res 024/2015": float,
        "Valor a favor para generador. Res 178/2015": float,
        "Valor a cargo para comercializador. Res 178/2015": float,
        "Valor a cargo para generador. Res 195/2015": float,
        "Valor a favor para generador. Res 195/2015": float,
        "Valor a favor para comercializador. Res 195/2015": float,
        "Valor a cargo para comercializador. Res 195/2015": float,
        "VALOR A CARGO PAGOS DE ENERGIA EXCEDENTARIA ($)": float,
        "VALOR A FAVOR POR ENERGIA EXCEDENTARIA ($)": float,
        "VC_RD resolución 011 de 2015": float,
        "VF_RD resolución 011 de 2015": float,
        "Valor a Favor delta ajuste RD": float,
        "Valor a Cargo delta ajuste RD": float,
        "VALOR A CARGO R026-2016 ($).": float,
        "VALOR A FAVOR R026-2016 ($).": float,
        "VALOR A FAVOR R029-2016 ($).": float,
        "RF039 resolución 039 de 2016": float,
        "RC039 resolución 039 de 2016": float,
        "Balance Final 029 de 2016": float,
        "Valor a cargo para comercializador. RES. 062 2013": float,
        "Valor a favor para generador. RES. 062 2013": float,
        "Valor del DE TIE. RES. 049 2018 (kWh)": float,
        "Valor del DE TIE. RES. 049 2018 ($)": float,
        "Magnitud desviación despacho. RES. 060 2019 (kWh)": float,
        "Valor Desviación Despacho. RES. 060 2019 ($)": float,
        "Magnitud desviación Redespacho. RES. 060 2019(kWh)": float,
        "Valor Desviación ReDespacho. RES. 060 2019 (kWh)": float,
        "Desviación Generación Variable. RES. 060 2019(kWh)": float,
        "Alivio desviaciones Res. CREG 060/2019 ($)": float,
        "VALOR PAGO AJUSTE RES. 140 2017 ($)": float,
        "VALOR COBRO AJUSTE RES. 140 2017 ($)": float,
        "VALOR PAGO EXCEDENTE RES. 140 2017 ($)": float,
        "VALOR COBRO FALTANTE RES. 140 2017 ($)": float,
    },
}


class AFAC(FileReader):
    def __init__(self):
        return super().__init__(afac_format.copy())


def afac_preprocess(filepath: pathlib.Path, item: FileItemInfo) -> pd.DataFrame:
    """
    AFAC: se publica un archivo por mes.
    versiones: TXR,TXF
    AGENTE:
     EPSC: el codigo del agente ##AQUI VA ENERBIT##
    """
    afac_reader = AFAC()
    total = afac_reader.read(filepath)
    if item.day is None:
        item.day = 1
    total["FECHA"] = f"{item.year:04d}-{item.month:02d}-{item.day:02d}"
    total["FECHA"] = pd.to_datetime(
        total["FECHA"],
        format="%Y-%m-%d",
    )
    total = total[total["AGENTE"] == item.agent]
    # retur_col = [
    #     "FECHA",
    #     "AGENTE",
    #     "PERDIDA REAL (kWh)",
    #     "DEMANDA REAL (kWh)",
    #     "GENERACION REAL (kWh)",
    #     "COMPRAS EN BOLSA (KWH)",
    #     "COMPRAS EN BOLSA ($)",
    #     "VENTAS EN BOLSA (KWH)",
    #     "VENTAS EN BOLSA ($)",
    #     "COMPRAS EN DESVIACION (KWH)",
    #     "COMPRAS EN DESVIACION ($)",
    #     "VENTAS EN DESVIACION (KWH)",
    #     "VENTAS EN DESVIACION ($)",
    #     "COMPRAS EN RECONCILIACION (KWH)",
    #     "COMPRAS EN RECONCILIACION ($)",
    #     "VENTAS EN RECONCILIACION (KWH)",
    #     "VENTAS EN RECONCILIACION ($)",
    #     "COMPRAS EN CONTRATOS (kWh)",
    #     "VENTAS EN CONTRATOS (kWh)",
    #     "COMPRAS ENERGIA EN BOLSA (KWH)",
    #     "COMPRAS ENERGIA EN BOLSA ($)",
    #     "VENTAS ENERGIA EN BOLSA (KWH)",
    #     "VENTAS ENERGIA EN BOLSA ($)",
    #     "VR CARGO POR CONFIABILIDAD ($)",
    #     "VD CARGO POR CONFIABILIDAD ($)",
    #     "NETO CXC ($)",
    #     "COMPRAS CARGO POR CONFIABILIDAD ($)",
    #     "VENTAS CARGO POR CONFIABILIDAD ($)",
    #     "COMPRAS EN BOLSA NACIONAL (KWH)",
    #     "COMPRAS EN BOLSA NACIONAL ($)",
    #     "VENTAS EN BOLSA NACIONAL (KWH)",
    #     "VENTAS EN BOLSA NACIONAL ($)",
    #     "COMPRAS EN BOLSA INTERNACIONAL (KWH)",
    #     "COMPRAS EN BOLSA INTERNACIONAL ($)",
    #     "VENTAS EN BOLSA INTERNACIONAL (KWH)",
    #     "VENTAS EN BOLSA INTERNACIONAL ($)",
    #     "SERVICIOS AGC ($)",
    #     "RESPONSABILIDAD COMERCIAL AGC (KWH)",
    #     "RESPONSABILIDAD COMERCIAL AGC ($)",
    #     "TOTAL COMPRAS ($)",
    #     "TOTAL VENTAS ($)",
    #     "VALOR A PAGAR POR SRPF ($)",
    #     "VALOR A RECIBIR POR SRPF ($)",
    #     "TOTAL RESTRICCIONES ($)",
    #     "RENTAS DE CONGESTION ($)",
    #     "RESTRICCIONES ALIVIADAS ($)",
    #     "VEBO (KWH)",
    #     "RENTAS DE CONGESTIÓN POR IMPORTACIÓN ($)",
    #     "DISTRIBUCIÓN SALDO NETO TIE EN MÉRITO ($)",
    #     "DISTRIBUCIÓN SALDO NETO TIE FUERA DE MÉRITO ($)",
    #     "COMPRAS BOLSA CON SALDO NETO TIE MÉRITO ($)",
    #     "RENDIMIENTOS FINANCIEROS POR EXPORTACIONES TIE($)",
    #     "ALIVIO POR CIOEF ($)",
    #     "COMPRAS NDC ($)",
    #     "VENTAS DESVIACIONES OEFH ($)",
    #     "COMPRAS DESVIACIONES OEFH ($)",
    #     "DEVOLUCION DINEROS DEL CARGO POR CONFIABILIDAD ($)",
    #     "COBRO DINERO CARGO POR CONFIABILIDAD ($)",
    #     "COMPRAS ARRANQUE Y PARADA ($)",
    #     "VENTAS ARRANQUE Y PARADA ($)",
    #     "Ventas por EEVE ($)",
    #     "Compras por EEVE ($)",
    #     "Restricciones por EEVE ($)",
    #     "Cobro uso respaldo($)",
    #     "Alivio restricciones RES 05/2010 ($)",
    #     "COMPRAS EN BOLSA TIES($)",
    #     " VENTAS EN BOLSA TIES ($)",
    #     " MAGNITUD EN Kwh  DE COMPRAS EN BOLSA DE TIES",
    #     " MAGNITUD EN kwh DE VENTAS EN BOLSA TIES",
    #     "ALIVIO POR EJECUCION DE GARANTIA ($)",
    #     "VALOR TOTAL EJECUCION DE GARANTIA ($)",
    #     "Alivio por VCSRCFVD($)",
    #     "VOEFV a cargo por la OEFV adquirida en la SRCFV($)",
    #     "VMOEFV a cargo al Margen del Precio MP SRCFV ($):",
    #     "Costo de Exportación ($)",
    #     "Total Costo de Exportación ($)",
    #     "Total de Generación Ideal en kWh del Agente",
    #     "Total de Holgura de AGC en kWh asignados al Agente",
    #     "Energía vendida y embalsada Asignada kWh",
    #     "VR Demanda Res 155/2014",
    #     "Alivio Asociado a la Resolución CREG 024/2015 en $",
    #     "Cobro Autogeneradores Res 024/2015",
    #     "Valor a favor para generador. Res 178/2015",
    #     "Valor a cargo para comercializador. Res 178/2015",
    #     "Valor a cargo para generador. Res 195/2015",
    #     "Valor a favor para generador. Res 195/2015",
    #     "Valor a favor para comercializador. Res 195/2015",
    #     "Valor a cargo para comercializador. Res 195/2015",
    #     "VALOR A CARGO PAGOS DE ENERGIA EXCEDENTARIA ($)",
    #     "VALOR A FAVOR POR ENERGIA EXCEDENTARIA ($)",
    #     "VC_RD resolución 011 de 2015",
    #     "VF_RD resolución 011 de 2015",
    #     "Valor a Favor delta ajuste RD",
    #     "Valor a Cargo delta ajuste RD",
    #     "VALOR A CARGO R026-2016 ($).",
    #     "VALOR A FAVOR R026-2016 ($).",
    #     "VALOR A FAVOR R029-2016 ($).",
    #     "RF039 resolución 039 de 2016",
    #     "RC039 resolución 039 de 2016",
    #     "Balance Final 029 de 2016",
    #     "Valor a cargo para comercializador. RES. 062 2013",
    #     "Valor a favor para generador. RES. 062 2013",
    #     "Valor del DE TIE. RES. 049 2018 (kWh)",
    #     "Valor del DE TIE. RES. 049 2018 ($)",
    #     "Magnitud desviación despacho. RES. 060 2019 (kWh)",
    #     "Valor Desviación Despacho. RES. 060 2019 ($)",
    #     "Magnitud desviación Redespacho. RES. 060 2019(kWh)",
    #     "Valor Desviación ReDespacho. RES. 060 2019 (kWh)",
    #     "Desviación Generación Variable. RES. 060 2019(kWh)",
    #     "Alivio desviaciones Res. CREG 060/2019 ($)",
    #     "VALOR PAGO AJUSTE RES. 140 2017 ($)",
    #     "VALOR COBRO AJUSTE RES. 140 2017 ($)",
    #     "VALOR PAGO EXCEDENTE RES. 140 2017 ($)"
    # ]
    return total



import datetime as dt
import ftplib
import logging
import os
import pathlib
from typing import Optional

import typer
from rich.progress import track

from asic import ASIC_FILE_EXTENSION_MAP
from asic.files import SupportedFiles
from asic.ftp.ftp import grab_file  # list_supported_files_in_location,
from asic.ftp.ftp import list_supported_files
from asic.publication import list_latest_published_versions

logger = logging.getLogger("eds")
logger.addHandler(logging.StreamHandler())

YEAR_MONTH_FORMATS = ["%Y-%m", "%Y%m"]
YEAR_MONTH_MATCH_ERROR_MESSAGE = f"Must match one of: {YEAR_MONTH_FORMATS}"
SUPPORTED_FILES = sorted([f.value.lower() for f in SupportedFiles])
SUPPORTED_FILES_ERROR_MESSAGE = f"Must match one of: {SUPPORTED_FILES}"
SUPPORTED_EXTENSIONS = [e.lower() for e in ASIC_FILE_EXTENSION_MAP.keys()]
SUPPORTED_EXTENSIONS_ERROR_MESSAGE = f"Must match one of: {SUPPORTED_EXTENSIONS}"
PUBLIC_SEARCHEABLE_LOCATIONS = [
    r"/PublicoK/SIC/COMERCIA/{year:04d}-{month:02d}/",
]
PRIVATE_SEARCHEABLE_LOCATIONS = [
    r"/UsuariosK/{agent}/Sic/COMERCIA/{year:04d}-{month:02d}/",
]


cli = typer.Typer(no_args_is_help=True)


@cli.callback()
def main(verbosity: int = typer.Option(0, "--verbosity", "-v", count=True)):
    logger.info(f"Verbosity level {verbosity}")
    match verbosity:
        case 0:
            logger.setLevel(logging.FATAL)
        case 1:
            logger.setLevel(logging.ERROR)
        case 2:
            logger.setLevel(logging.WARNING)
        case 3:
            logger.setLevel(logging.INFO)
        case 4:
            logger.setLevel(logging.DEBUG)
        case _:
            logger.setLevel(logging.DEBUG)


def parse_month(month: str) -> dt.date:

    for f in YEAR_MONTH_FORMATS:
        try:
            value = dt.datetime.strptime(month, f).date()
            break
        except ValueError:
            continue
    else:
        raise typer.BadParameter(YEAR_MONTH_MATCH_ERROR_MESSAGE)

    return value


def validate_file(file_code: str) -> str:
    if not file_code.lower() in SUPPORTED_FILES:
        raise typer.BadParameter(SUPPORTED_FILES_ERROR_MESSAGE)
    return file_code


def validate_version(version: str) -> str:
    if not version.lower() in SUPPORTED_EXTENSIONS:
        raise typer.BadParameter(SUPPORTED_EXTENSIONS_ERROR_MESSAGE)
    return version


def months_callback(values: list[str]) -> list[dt.date]:
    months = sorted(list(set([parse_month(v) for v in values])), reverse=True)
    return months


def files_callback(values: list[str]) -> list[str]:
    files = sorted(
        list(set([validate_file(v) for v in values])),
        reverse=True,
    )

    return files


def extensions_callback(values: list[str]) -> list[str]:
    extensions = list(set([validate_version(v) for v in values]))

    return extensions


@cli.command()
def pubs(
    days_old: int = typer.Option(None),
    after: dt.datetime = typer.Option(None),
    include_daily: bool = typer.Option(False),
):
    """
    Check latest published settlements in asic's website.
    """
    message = "Listing latest published settlements by ASIC"
    if days_old is not None:
        message = message + f" in the last {days_old} days"
        published_after = dt.datetime.combine(
            dt.date.today(), dt.datetime.min.time()
        ) + dt.timedelta(days=-days_old)
    elif after is not None:
        message = message + f" afte {after}"
        published_after = after
    else:
        published_after = None
    logger.info(message)
    latest_publications = list_latest_published_versions(published_after, include_daily)
    for pub in latest_publications:
        typer.echo(
            f"{pub.month:%Y-%m}:{pub.version.upper()} -- published: {pub.published_at:%Y-%m-%d}"
        )


@cli.command("list")
def list_files(
    months: list[str] = typer.Option(
        ...,
        "--month",
        callback=months_callback,
        help=YEAR_MONTH_MATCH_ERROR_MESSAGE,
    ),
    agent: Optional[str] = typer.Option(
        None, help="Agent's asic code, required for private files"
    ),
    files: Optional[list[str]] = typer.Option(
        None, "--file", callback=files_callback, help=SUPPORTED_FILES_ERROR_MESSAGE
    ),
    extensions: Optional[list[str]] = typer.Option(
        None,
        "--version",
        callback=extensions_callback,
        help=SUPPORTED_EXTENSIONS_ERROR_MESSAGE,
    ),
    ftp_host: str = typer.Argument(..., envvar="ASIC_FTP_HOST"),
    ftp_user: str = typer.Argument(..., envvar="ASIC_FTP_USER"),
    ftp_password: str = typer.Argument(..., envvar="ASIC_FTP_PASSWORD"),
    # asic_raw_container_name: str = typer.Argument(
    #     "asic-raw", envvar="ASIC_RAW_CONTAINER_NAME"
    # ),
    # asic_processed_container_name: str = typer.Argument(
    #     "asic", envvar="ASIC_PROCESSED_CONTAINER_NAME"
    # ),
):
    """
    List files from asic's ftp server.

    FTP authentication info should be  provided as environment variables (ASIC_FTP_*)
    """
    logger.info(
        "Listing files for"
        f" agent: {agent}"
        f" months: {months}"
        f" extensions: {extensions}"
        f" files: {files}"
    )
    if not extensions:
        extensions = [None]
    if not files:
        files = SUPPORTED_FILES

    locations = PUBLIC_SEARCHEABLE_LOCATIONS
    if agent is not None:
        locations.extend(PRIVATE_SEARCHEABLE_LOCATIONS)

    with ftplib.FTP(
        host=ftp_host,
        user=ftp_user,
        passwd=ftp_password,
    ) as ftp_aux:
        file_list = list_supported_files(
            ftp_aux,
            agent=agent,
            months=months,
            extensions=extensions,
            files=files,
            locations=locations,
        )
    for f in file_list:
        typer.echo(f)


@cli.command()
def download(
    months: list[str] = typer.Option(
        ...,
        "--month",
        callback=months_callback,
        help=YEAR_MONTH_MATCH_ERROR_MESSAGE,
    ),
    agent: Optional[str] = typer.Option(
        None, help="Agent's asic code, required for private files"
    ),
    files: Optional[list[str]] = typer.Option(
        None, "--file", callback=files_callback, help=SUPPORTED_FILES_ERROR_MESSAGE
    ),
    extensions: Optional[list[str]] = typer.Option(
        None,
        "--version",
        callback=extensions_callback,
        help=SUPPORTED_EXTENSIONS_ERROR_MESSAGE,
    ),
    destination: pathlib.Path = typer.Argument(...),
    ftp_host: str = typer.Argument(..., envvar="ASIC_FTP_HOST"),
    ftp_user: str = typer.Argument(..., envvar="ASIC_FTP_USER"),
    ftp_password: str = typer.Argument(..., envvar="ASIC_FTP_PASSWORD"),
    # asic_raw_container_name: str = typer.Argument(
    #     "asic-raw", envvar="ASIC_RAW_CONTAINER_NAME"
    # ),
    # asic_processed_container_name: str = typer.Argument(
    #     "asic", envvar="ASIC_PROCESSED_CONTAINER_NAME"
    # ),
):
    """
    Download files from asic's ftp server to local DESTINATION folder.

    FTP authentication info should be  provided as environment variables (ASIC_FTP_*)
    """
    if not extensions:
        extensions = [None]
    if not files:
        files = SUPPORTED_FILES

    locations = PUBLIC_SEARCHEABLE_LOCATIONS
    if agent is not None:
        locations.extend(PRIVATE_SEARCHEABLE_LOCATIONS)

    with ftplib.FTP(
        host=ftp_host,
        user=ftp_user,
        passwd=ftp_password,
    ) as ftp_aux:
        file_list = list_supported_files(
            ftp_aux,
            agent=agent,
            months=months,
            extensions=extensions,
            files=files,
            locations=locations,
        )

    logger.info(f"Total files to download: {len(file_list)}")

    with ftplib.FTP(
        host=ftp_host,
        user=ftp_user,
        passwd=ftp_password,
    ) as ftp_aux:
        for f in track(file_list, description="Dowloading files..."):
            remote = f
            local = destination / str(f)[1:]  # hack to remove root anchor
            os.makedirs(local.parent, exist_ok=True)
            logger.info(f"Downloading {remote} to {local}")
            grab_file(ftp_aux, remote, local)



import logging
import pathlib

# Third party imports
import pandas as pd

# Local application imports
from asic.reader import FileReader

from ..metadata import FileItemInfo

logger = logging.getLogger(__name__)

trsd_format = {
    "type": "csv",
    "sep": ";",
    "encoding": "cp1252",
    "dt_fields": {},
    "dtype": {
        "CODIGO": str,
        "CONTENIDO": str,
        "HORA 01": float,
        "HORA 02": float,
        "HORA 03": float,
        "HORA 04": float,
        "HORA 05": float,
        "HORA 06": float,
        "HORA 07": float,
        "HORA 08": float,
        "HORA 09": float,
        "HORA 10": float,
        "HORA 11": float,
        "HORA 12": float,
        "HORA 13": float,
        "HORA 14": float,
        "HORA 15": float,
        "HORA 16": float,
        "HORA 17": float,
        "HORA 18": float,
        "HORA 19": float,
        "HORA 20": float,
        "HORA 21": float,
        "HORA 22": float,
        "HORA 23": float,
        "HORA 24": float,
    },
}


class TRSD(FileReader):
    def __init__(self):
        return super().__init__(trsd_format.copy())


def trsd_preprocess(filepath: pathlib.Path, item: FileItemInfo) -> pd.DataFrame:
    """
    trsd: se publica un archivo por día.
    versiones: TX1, TX2, TXR y TXF
    CODIGO:
      PBNA: Precio de bolsa nacional
    """
    trsd_reader = TRSD()
    total = trsd_reader.read(filepath)
    total["FECHA"] = f"{item.year:04d}-{item.month:02d}-{item.day:02d}"
    total = total[total["CODIGO"] == "PBNA"]
    total["FECHA"] = pd.to_datetime(
        total["FECHA"],
        format="%Y-%m-%d",
    )
    total = total.set_index(["CODIGO", "CONTENIDO", "FECHA"]).stack().reset_index()
    total = total.rename(columns={"level_3": "NOMBRE HORA", 0: "PRECIO"})
    total["HORA"] = (total["NOMBRE HORA"].str.slice(start=-2)).astype(int) - 1
    total["HORA"] = pd.to_timedelta(total["HORA"], unit="h")
    total["FECHA_HORA"] = total["FECHA"] + total["HORA"]
    total["HORA"] = total["FECHA_HORA"].dt.strftime("%H:%M:%S")
    total["FECHA"] = total["FECHA"].dt.date
    ret_cols = ["FECHA_HORA", "CODIGO", "CONTENIDO", "PRECIO"]
    return total[ret_cols]



# Standard library imports
import logging
from io import StringIO
from pathlib import Path
from typing import Union

# Third party imports
import pandas as pd

# Local application imports
pass

logger = logging.getLogger(__name__)


class FileReader:
    """Base Class for file readers."""

    def __init__(self, file_def: dict) -> None:
        """Create FileReader with file definition."""
        logger.debug(f"Creating FileReader with file definition:\n{file_def}")
        self.file_def = file_def

    def read(self, target: Union[str, Path, StringIO]) -> pd.DataFrame:
        """Reads DataFrame from target."""
        _def = self.file_def
        file_type = _def.pop("type", None)
        dt_fields = _def.pop("dt_fields", None)

        logger.debug(f"Reading as {file_type}: {target}")

        if file_type in ["csv", "txt", "tsv"]:
            res = pd.read_csv(target, **_def)

        elif file_type in ["xls", "xlsx"]:
            if "sheet_name" in _def:
                res = pd.read_excel(str(target), **_def)
            else:
                raise ValueError(
                    "Missing 'sheet_name' in file_def. Can only read one Excel sheet structure per FileReader"
                )

        else:
            raise ValueError(f"Not supported 'type': '{file_type}'")

        if dt_fields is not None:
            res = self._date_converter(res, dt_fields)

        logger.info(f"Read {res.shape[0]} rows and {res.shape[1]} columns")
        # logger.debug(f"Dtypes of result DataFrame:\n{res.dtypes}")
        return res

    def _date_converter(self, df: pd.DataFrame, dt_fields: dict) -> pd.DataFrame:
        for c in dt_fields:
            df[c] = pd.to_datetime(df[c], format=dt_fields[c].get("format"))

        return df
